module OrientDb
  class LearningObjectRepository < Base
    include OrientDb::Methods::EdgeMethods
    include RepositoriesProxy


    ##
    # Call this methods for increment the views number of a +learning_object+
    # You are creating the 'views' relation between an +user+ and +learning_object+
    # This relation is created only once.
    #
    # Nothing is returned
    def increment_views(user, learning_object)
      values = {date: sanitize_orientdb_values(Time.now)}

      # If doesn't exist edges from user and learning object, create it and increment attribute
      if !edge_exists? "Views", user.rid, learning_object.id
        create_and_set_edge "Views", user.rid, learning_object.id, values
        learning_object.views = learning_object.views + 1
      # If exists, update the date property of view
      else
        edge = get_edges "Views", "in", learning_object.id
        edge = edge.first

        puts edge
        set_values_edge edge["@rid"], values
      end
    end

    def download(user, learning_object)
      download_edge_class = 'Downloads'
      values = {date: sanitize_orientdb_values(Time.now)}

      create_and_set_edge download_edge_class, user.rid, learning_object.id, values
      learning_object.downloads = learning_object.downloads + 1
    end

    ##
    # Call this method checks if the +user+ has a like relation with +learning_object+
    # This methods returns a boolean
    def liked?(user, learning_object)
      edge_exists? 'Likes', user.rid, learning_object.id
    end

    ##
    # Call this method to create a 'like' relation between an user and a learning object.
    # If the user already has the like relation, then nothing happens.
    #
    # If you want raise an exeception, use the LearningObjectRepository::like! method.
    #
    # Usage:
    #   repository.for(:learning_objects).like current_user, @learning_object
    def like(user, learning_object)
      like_edge_class = 'Likes'

      # If not liked, like and set date
      if !liked? user, learning_object
        # The properties of edge. In this case, only date for edge 'Likes'
        values = {date: sanitize_orientdb_values(Time.now)}

        create_and_set_edge like_edge_class, user.rid, learning_object.id, values
        learning_object.likes = learning_object.likes + 1
      end
    end

    ##
    # Call this method to create a 'like' relation between an user and a learning object.
    # If the +user+ already liked the +learning_object+ throw an exception.
    def like!(user, learning_object)
      if liked? user, learning_object
        raise 'The user already likes this object.'
      end
      like user, learning_object
    end

    ##
    # This methods destroy a 'like' relation between +user+ and +learning_object+
    # If the +user+ hasn't liked the +learning_object+ nothing happens
    # If you want raise an exeception, use the LearningObjectRepository::dislike! method.
    def dislike(user, learning_object)
      like_edge_class = 'Likes'
      if liked? user, learning_object
        destroy_edge like_edge_class, user.rid, learning_object.id
        learning_object.likes = learning_object.likes - 1
        true
      end
    end

    ##
    # This methods destroy a 'like' relation between +user+ and +learning_object+
    # If the +user+ hasn't liked the +learning_object+ throw an exception.
    def dislike!(user, learning_object)
      if liked? user, learning_object
        dislike user, learning_object
      end
      raise 'The user hasn`t likes this object.'
    end

    #it could perhaps be in the GenericMethods, if we'd like to extend to User counting.
    def get_number_of_non_visualised
      (connection.query("SELECT COUNT(*) FROM LearningObject WHERE in('Views').size() = 0"))[0]["COUNT"].to_i
    end

    #get the ten most visualised. It's probably making too many useless accesses to orientDB...
    def get_most_visualised
      result = connection.query("SELECT FROM (SELECT @rid,in('Views').size() AS views FROM LearningObject) ORDER BY views DESC LIMIT 10")
      most_visualised = []
      result.each do |result|
        most_visualised.push(get_by_rid(result["rid"]))
      end
      build_objects most_visualised
    end

    def get_publisher(learning_object)
      result = get_edges_end("PublishedBy", "out", learning_object.id)
      result = result.first

      unless result.empty?
        if result["@class"] == "Institution"
          institution_repository.build_object(result)
        else
          user_repository.build_object(result)
        end
      end
    end

    # Usage:
    #   learning_object = repository.for(:learning_objects).get_by_dspace_id 123
    #
    def get_by_dspace_id(id_dspace)
      result = select_by_property(odb_class, "id_dspace", id_dspace)
      build_object result.first
    end

    ##
    # This method return all learning objects of +user+ scope.
    #
    # No users have objects.
    # Learning Object submission was not implemented yet.
    # For now, an empty array is returned.
    def all(user)
      #result = select_by_property(odb_class, "p_id", user.id)
      #build_objects result
      []
    end

    def all_by_score_from_offset_to_limit(offset, limit)
      objects_hash = connection.query "SELECT FROM LearningObject ORDER BY score DESC LIMIT #{limit} SKIP #{offset}", limit: limit
      objects = build_objects(objects_hash) || []
    end

    def get_subjects(learning_object)
      result = get_edges_end("IsAbout", "out", learning_object.id)
      subject_repository.build_objects(result)
    end

    def get_attributes(learning_object)
      result = get_edges_end("HasAttr", "out", learning_object.id)
      attribute_repository.build_objects(result)
    end

    def create_relations(learning_object)
      edges = []
      unless learning_object.subjects.empty?
        edges << create_edges_from_array("IsAbout", learning_object.id, learning_object.subjects, true)
      end
      unless learning_object.attributes.empty?
        edges << create_edges_from_array("HasAttr", learning_object.id, learning_object.attributes, true)
      end

      if !learning_object.publisher.nil? && !edge_exists?("PublishedBy", learning_object.id, learning_object.publisher.id)
        edges << create_edge("PublishedBy", learning_object.id, learning_object.publisher.id)
      end
      edges.flatten
    end

    def save(learning_object)
      if learning_object.valid?
        connection.command "INSERT INTO LearningObject CONTENT #{learning_object.to_json}"
      end
    end

    def update_property(learning_object, property, value)
      if accepted_properties.include? property
        connection.command "UPDATE LearningObject SET #{property}='#{value}', last_modified='#{Time.now()}' WHERE @rid = #{learning_object.id}"
      end
    end

    def types
      Rails.cache.fetch("learning_object/types", expires_in: 1.days) do
        query = "select type from LearningObject GROUP BY type"
        types_hash = connection.query query, limit: 100
        # return array with types
        types_hash.map do |e|
          e['type']
        end
      end
    end

    ##
    # To create "index:learningobject_search" on OrientDB, use the following command:
    #   CREATE INDEX learningobject_search
    #   ON LearningObject (name, description)
    #   FULLTEXT ENGINE LUCENE
    #   METADATA {"analyzer":"org.apache.lucene.analysis.br.BrazilianAnalyzer"}
    def search(params)
      # mount cache_key with params
      valid_params = %w(query order subject type school_level year)
      cache_key = "search_result/"
      valid_params.each do |param|
        if params[param].class == Array
          cache_key += params[param].join('-')
        else
          cache_key += params[param].to_s
        end
        cache_key += '/' unless (param == valid_params.last)
      end

      # get results in cache or search
      Rails.cache.fetch(cache_key, expires_in: 10.minutes) do
        qry = params[:query]

        order = params[:order]
        order = case order
                  when 'author'
                    'author'
                  when 'publicationasc'
                    'published_at'
                  when 'publicationdesc'
                    'published_at DESC'
                  when 'title'
                    'name ASC'
                  else
                    'score DESC'
                end

        query = "SELECT FROM LearningObject WHERE [name, description, author] LUCENE '#{qry}'"
        query += " AND out('IsAbout') CONTAINS (name in ['" + params[:subject].join("','") + "'])" unless params[:subject].blank?
        unless params[:school_level].blank?
          params[:school_level] << 'Ensino Fundamental Inicial' << 'Ensino Fundamental Final' if params[:school_level].include? 'Ensino Fundamental'
          query += " AND out('IsAbout') CONTAINS (name in ['" + params[:school_level].join("','") + "'])"
        end
        query = "SELECT * FROM (" + query + ") WHERE " + fetch_types(params[:type]) unless params[:type].blank?
        query = "SELECT * FROM (" + query + ") WHERE " + fetch_year(params[:year]) unless params[:year].blank?
        query = "SELECT * FROM (" + query + ") ORDER BY #{order}"
        query = "SELECT @rid.asString(), last_modified FROM (" + query + ")"
        connection.query query, limit: 10000
      end
    end

    def build_object(args={})
      LearningObjectBuilder.build_from_orientdb args
    end

    def build_hash(object)
      hash = super(object)
      hash.delete("likes")
      hash.delete("views")
      hash.delete("downloads")
      hash.delete("subjects")
      hash.delete("attributes")

      hash["metadata"] = hash["metadata"].to_json
      hash
    end

    def fetch_year(years)
      if years.size() == 1
        year = years[0]
        qry = "created_at.format('yyyy') = '#{year}'"
      elsif years.size() == 2
        begin_year = years[0]
        end_year = years[1]
        qry = "created_at.format('yyyy') >= '#{begin_year}' AND created_at.format('yyyy') <= '#{end_year}'"
      end
    end

    def fetch_types(types)
      qry = ''
      types.each do |type|
        qry = qry + "type = '#{type}' " if type == types.first
        qry = qry + "OR type = '#{type}' " if type != types.first
      end
      qry
    end

    def count_all
      Rails.cache.fetch('learning_object/count_all', expires_in: 6.hours) do
        result = connection.query 'SELECT COUNT(*) as count FROM LearningObject', limit: 1
        result[0]["count"]
      end
    end

    def count_likes(learning_object)
      get_in_edges_count "Likes", learning_object.id
    end

    def count_views(learning_object)
      get_in_edges_count "Views", learning_object.id
    end

    def count_downloads(learning_object)
      get_in_edges_count "Downloads", learning_object.id
    end

    def max_likes
      Rails.cache.fetch('max_likes', expires_in: 6.hours) do
        get_max_from_edge("Likes","in")
      end
    end

    def max_views
      Rails.cache.fetch('max_views', expires_in: 6.hours) do
        get_max_from_edge("Views","in")
      end
    end

    def max_downloads
      Rails.cache.fetch('max_downloads', expires_in: 6.hours) do
        get_max_from_edge("Downloads","in")
      end
    end

    def get_max_from_edge(edge_class, type)
      max = connection.command "SELECT max(#{type}('#{edge_class}').size()) FROM LearningObject"
      max["result"][0]["max"]
    end

    private

    def accepted_properties
      ['thumbnail', 'score']
    end

    def odb_class
      "LearningObject"
    end

  end
end
