namespace :orientdb do
  desc "Orient DB integration tasks"

  task migrate: :environment do
    desc "Migrate OrientDB schema"

    migrations = OrientDb::Migrations.new(OrientDb::Client.instance)
    migrations.run
  end

  task drop: :environment do
    desc "Drop OrientDB schema"

    migrations = OrientDb::Migrations.new(OrientDb::Client.instance)
    migrations.drop
  end

  task create_learning_object_relations: :environment do
    desc "Create LearningObject relations based on its metadata"

    include RepositoriesProxy

    # Quantity of LearningObjects fetched on each iteration
    limit = 1000
    # Starting point from where LearningObjects will be fetched
    offset = 0

    loop do
      puts " --> creating LearningObjects realations from #{offset} to #{offset+limit}"

      begin
        # Get LearningObjects from OrientDB (from offset to offset+limit)
        learning_objects = learning_object_repository.all_from_offset_to_limit(offset,limit)
      rescue => e
        puts e
        puts e.backtrace
        # Sleeps for a while to wait database's recovery
        sleep(30.seconds)
        # Goes to next iteration to retry
        next
      else
        # Terminate loop if there are no more LearningObjects
        break if learning_objects.empty?

        learning_objects.each do |lo|
          puts "LO ID ==== "+lo.id.to_s
          # Get metadata as a hash with metadata name as key
          metadata = get_unique_matadata_keys(lo)
          # Get subjects
          subjects = metadata["dc.subject.category"]
          # If subjects is nil, set as empty array
          subjects ||= []
          # Remove subjects from the metadata hash, so they wont be inserted as Attributes
          metadata.delete("dc.subject.category")
          # Set LearningObject relations attributes so it wont perform a search in the database when accessed
          lo.subjects=[]
          lo.attributes=[]
          # Remove nil values from subjects array
          subjects.compact!
          subjects.each do |name|
            # Look for subject with the same name
            subject = subject_repository.find_by_name(name)
            if subject.nil?
              # Create a Subject with this name if it doesnt exist
              subject = Subject.new(:name => name)
              subject_repository.create(subject)
            end
            # Add subject on the LearningObject list
            lo.subjects << subject
          end

          metadata.each do |key, values|
            # Remove nil values from values array
            values.compact!
            values.each do |v|
              # Remove characters that might throw exceptions in OrientDB
              v = v.delete('\\').delete('[').delete(']')
              # Look for attribute with the same key and value
              attribute = attribute_repository.find_by_key_and_value(key,v)
              if attribute.nil?
                # Create an Attribute with this key and value if it doesnt exist
                attribute = Attribute.new(:key => key, :value => v)
                attribute_repository.create(attribute)
              end
              # Add attribute on the LearningObject list
              lo.attributes << attribute
            end
          end
          
          # Create LearningObject relations that were added to its attrs
          learning_object_repository.create_relations(lo)
        end
        # Increase offset for next loop
        offset += limit
      end
    end

  end

  # For each unique key on LearningObject metadata, get all its values and add on a new hash
  def get_unique_matadata_keys(lo)
    hash = {}
    lo.metadata.each do |m|
      unless hash.has_key?(m['key'])
        hash[m['key']] = lo.get_metadata_values_of(m['key'])
      end
    end
    hash
  end

end
